/***************************************************************************\

  File........: avr_usart.c
  Author(s)...: Gerhard Brnner
  Target(s)...: AVR_MEGA8
  Compiler....: WIN_AVR
  Description.: Routines for the internal hardware USART.
  				Sending - without interrupt support
				Recveiving - with interrput support
  				Including command interface for received commands

  Revisions...: V2.1
  2009.11.11  - removed type byte_t
  2009.02.01  - added sub-functions for receive-buffer and receive-interrupt
  2008.07.04  - added request for defined USART0_RX_vect ...

\***************************************************************************/


#define AVR_USART_C

/* Include-Dateien
****************************************************************************/
	#include "avr_usart.h"

/* Funktions-Prototypen
****************************************************************************/
	void ringbuffer_t_init( ringbuffer_t *buf );
	int8_t ringbuffer_add(ringbuffer_t *buf, uint8_t data);
	int8_t ringbuffer_read(ringbuffer_t *buf, uint8_t *data);

/* Interne Variablen
****************************************************************************/
	static ringbuffer_t rcv_buffer;


/*********************************************************************************************\
*
*
* 	Funktionen
*
*
\*********************************************************************************************/

/***********************************************************************************
	FUNKTION:	init_usart(...)
	Diese Funktion dient der Initialsierung der UART-Schnittstelle.
	Zur Vermeidung von berschssigen Code erfolgt die Parametrierung durch defines
	im Header-File
************************************************************************************/

/* Interface parameter to adapt
****************************************************************************/

void init_usart_9600_8_1_N_Asyn( void )
{

/* USART Control Register A - Settings
************************************************************************/

	/* Register A wird nur beim Empfang und Versenden bentigt.*/

/* USART Control Register B - Settings
	Bit 7 		6 		5 		4 		3 		2 		1 		0
		RXCIE 	TXCIE 	UDRIE 	RXEN 	TXEN 	UCSZ2 	RXB8 	TXB8
	Bit 7 - RXCIE: RX Complete Interrupt Enable
	Bit 6 - TXCIE: TX Complete Interrupt Enable
	Bit 5 - UDRIE: USART Data Register Empty Interrupt Enable
	Bit 4 - RXEN: Receiver Enable
	Bit 3 - TXEN: Transmitter Enable
	Bit 2 - UCSZ2: Character Size wird bei Verwendung von 9 Datenbits gesetzt
	Bit 1 - RXB8: Receive Data Bit - contains 9th databit
	Bit 0 - TXB8: Transmit Data Bit - contains 9th databit
************************************************************************/
	UCSRB=0x00;
	UCSRB |= ( 1 << TXEN )    /*UART TX / Sendung einschalten */
		  |  ( 1 << RXEN )    /*UART RX / Empfang einschalten*/
	      |  ( 1 << RXCIE );   /*Bit RXCIE - RX Complete Interrupt Enable*/
		/*|  ( 1 << TXCIE );	 Bit TXCIE (TX Complete Interrupt Enable)*/


/* USART Control Register C - Settings
	Bit 7 		6 		5 		4 		3 		2 		1 		0
		URSEL 	UMSEL 	UPM1 	UPM0 	USBS 	UCSZ1 	UCSZ0 	UCPOL
	The UCSRC Register shares the same I/O location as the UBRRH Register.
	Bit 7 URSEL: Register Select
		This bit selects between accessing the UCSRC or the UBRRH Register.
		It is read as one when reading UCSRC. The URSEL must be one when writing the UCSRC.
	Bit 6  UMSEL: USART Mode Select
	This bit selects between Asynchronous and Synchronous mode of operation.
************************************************************************/
	UCSRC=0x00;
	/* Async./sync mode */
		UCSRC |= (1<<UMSEL); /* 0 - asynchron / 1 - synchrom */
	/* Paritt */
		UCSRC |= (0<<UPM0)|(0<<UPM1); /* none*/
			/* (0<<UPM0)|(1<<UPM1);  even */
			/* (1<<UPM0)|(1<<UPM1);  odd */
	/* Stoppbits */
		UCSRC |= (0<<USBS); /* 0= 1 Stoppbit; 1=2 Stoppbits */
	/* Datenbits */
		UCSRC |= (1<<UCSZ1)|(1<<UCSZ0);
		UCSRB |= (0<<UCSZ2); /*1 in case of 9 databit*/
		/*UCSZ2(in UCSRB), UCSZ1, UCSZ0	Datenbits
			0		0		0	5
			0		0		1	6
			0		1		0	7
			0		1		1	8
			1		1		1	9
		*/
	/* UCPOL	*/
		UCSRC |= (0<<UCPOL); /* usart_mode = asynchronous */

/* Baudrate einstellen
************************************************************************/
	UBRRH = (uint8_t)(UART_UBRR_CALC(9600)>>8); /*9600UL 19200UL 57600UL 115200UL*/
	UBRRL = (uint8_t)UART_UBRR_CALC(9600);
	ringbuffer_t_init(&rcv_buffer);
} /* Ende Funktion: void init_usart( void ) */


/***********************************************************************************\

	Send Routines

\***********************************************************************************/
		/* Zeichen senden */
		void Usart_putc(uint8_t data)
		{
			while ((UCSRA & 1<<UDRE) == 0); /* warten bis Senden moeglich */
			UDR = data; /* schreibt das Zeichen x auf die Schnittstelle */
		}

		/* Senden von Strings*/
		void Usart_puts (char *s)
		{
			while (*s)
			{ /* so lange *s != '\0' also ungleich dem "String-Endezeichen" */
				Usart_putc((uint8_t) *s);
				s++;
			}
		}

/***********************************************************************************\

	Receive Routines

\***********************************************************************************/

	/************************************************************\
		Interrupt-Routine: ISR (USART0_RX_vect)
	\************************************************************/

		ISR (USART_RXC_vect) 	/* INTERRUPT - USART, Rx Complete */
		{
			register uint8_t sreg, rcv_byte;
			sreg=SREG; 			/* SREG sichern */
			rcv_byte = UDR;
			ringbuffer_add(&rcv_buffer, rcv_byte);
			SREG=sreg;  		/* SREG wiederherstellen */
		}


/***********************************************************************************\






	Helping Routines






\***********************************************************************************/

/*********************************************************************************************\
	Function :	ringbuffer_t_init
	Parameter:	buffer_length_in_byte - number of byte required for the buffer.
	RetVal:	 	void

	initializes the buffer structure and allocates the required memory
\*********************************************************************************************/
void ringbuffer_t_init(ringbuffer_t *buf ){
int8_t cnt=_MAX_BUFFER_LEN-1;

    buf->WritePos =  buf->Overflow_cnt = buf->ReadPos  =0;
	for(;cnt>=0;cnt--) buf->Data[cnt]=0;
} /*int8_t ringbuffer_t_init( uint16_t buffer_length_in_byte);*/

/*********************************************************************************************\
	Function :	ringbuffer_add
	Parameter:
		ringbuffer_t 	*buf	-	Pointer auf Puffervariable
		uint8_t 		data	-	zu schreibendes Byte
	RetVal:
		0	-	okay
		-1	-	buffer overflow

	writes byte into the buffer structure
\*********************************************************************************************/
int8_t ringbuffer_add(ringbuffer_t *buf, uint8_t data) {
	buf->Data[buf->WritePos++]=data;
	buf->WritePos&=~_MAX_BUFFER_LEN;
	if (buf->WritePos==buf->ReadPos){
	    buf->Overflow_cnt++;
		return(-1);
	}
	return(0);
} /* int8_t ringbuffer_add(ringbuffer_t *buf, uint8_t data) */

/*********************************************************************************************\
	Function :	ringbuffer_read
	Parameter:
		ringbuffer_t 	*buf	-	Pointer auf Puffervariable
		uint8_t 		*data	-	Pointer auf Speicherstelle in der das gelesene Byte
									ausgegeben werden soll
	RetVal:
		=0	-	nothing to read
		>0	-	amount of bytes read
		<0	-	error

	read byte from the buffer structure
\*********************************************************************************************/
int8_t ringbuffer_read(ringbuffer_t *buf, uint8_t *data){
	/* ReadPos = WritePos >> nothing to read */
	if(buf->ReadPos!=buf->WritePos){
		*data=buf->Data[buf->ReadPos++];
		buf->ReadPos&=~_MAX_BUFFER_LEN;
		return(1);
	}

	return(0);
} /* int8_t ringbuffer_read */

/*********************************************************************************************\
	Function :	ringbuffer_read_string
	Parameter:
		ringbuffer_t 	*buf	-	Pointer auf Puffervariable
		uint8_t 		*data	-	Pointer auf Speicherstelle in der das gelesene Byte
									ausgegeben werden soll
	RetVal:
		=0	-	nothing to read
		>0	-	amount of bytes read
		<0	-	error

	read byte from the buffer structure
\*********************************************************************************************/
int8_t ringbuffer_read_string(ringbuffer_t *buf, uint8_t *data){
register int8_t ret_val=0;
	/* ReadPos = WritePos >> nothing to read */
	while(buf->ReadPos!=buf->WritePos){
		*data=buf->Data[buf->ReadPos++];
		buf->ReadPos&=~_MAX_BUFFER_LEN;
		data++;
		ret_val++;
	}

	return(ret_val);
} /* int8_t ringbuffer_read */



/***********************************************************************************\




	Command Interface





\***********************************************************************************/
uint8_t usart_read_buffer_string(uint8_t *data){
	register uint8_t cnt=0;
	cnt = ringbuffer_read_string(&rcv_buffer, data);
	data[cnt]=0; /* Set EOS */
	return(cnt);
} /* int8_t ringbuffer_read */


#undef AVR_USART_C
