/***************************************************************************/
/**
	@file		cmd_interface.c
	@brief		Simple implementation of an command-interface
	@brief		Target/Compiler: AVR_MEGA / GCC(WinAVR)
	@author		Gerhard Bruenner
	@version	V2.01

	@date		2009.11.15  - added function cmd_interface_add_cmd and Backspace-Function - GB
	@date		2009.10.31  - changed the way how commandline is interpret - GB
	@date		2009.10.21  - small updates   - GB
	@date		2009.02.01  - Created         - GB

	@bug		none that I am aware so far

\***************************************************************************/

#define   CMD_INTERFACE_C

/* Include-Dateien
   ************************************************/
	#include "cmd_interface.h"
	#include <stdlib.h>
	#include <string.h>
	#include "string_addon.h"

/* globale int Variablen
   ************************************************/
			char   	cmd_line[_CMD_INTERFACE_MAX_LENGTH+1];
	static 	int8_t	cmd_line_act_length=0;
	static 	char 	cmd_line_split_parts[_CMD_INTERFACE_MAX_NO_OPERANDS][_CMD_INTERFACE_MAX_OPERAND_LENGTH];
	volatile 		cmd_interface_command_table_t cmd_interface_table[_CMD_INTERFACE_NO_OF_COMMANDS];

/* externe Variablen
   ************************************************/


/* internal Funktions-Prototypen
****************************************************************************/
	int8_t execute_command();
	int8_t split_command_line();

/**
 	@brief		initialisation routine for command interface variables
 	@param		none
 	@return		void
 */
void cmd_interface_init(){
	register uint8_t ui_cnt=0;
	for(ui_cnt=0; _CMD_INTERFACE_NO_OF_COMMANDS>ui_cnt; ui_cnt++) {
		cmd_interface_table[ui_cnt].cmd=0;
		cmd_interface_table[ui_cnt].cmd_func_ptr=NULL;
	}
}/* void cmd_interface_init()*/

/**
 	@brief		adding a command to internal command - table

 	@param		char *cmd_string 		- String representing the command
 	@param		uint8_t (*cmd_func_ptr) - pointer to function which shall be executed if the command has been found

 	@return		 0 - all fine
 				-1 - command-table is full
 	@details	This function shall add a given string and function-pointer to the internal command-table.
 				The string in the command-table is used from the internal function to determine whether
 				a command is known. The function where the function-pointer is pointing to is called in
 				the case the entered command matches to the given string.
 				The function MUST have the following interface-declaration....
 				uint8_t		cmd_func (char *op1, char *op2, char *op3, char *op4)
 */
uint8_t cmd_interface_add_cmd(char (*cmd_string), uint8_t (*cmd_func_ptr)() ){
	static uint8_t ui_static_cmd_counter=0;
	uint8_t ret_val = -1; /* ERROR: -1 >> Command-Table is full */

	if(_CMD_INTERFACE_NO_OF_COMMANDS>=ui_static_cmd_counter){ 					/* there are still commands left to be set */
		trim(&cmd_string[0],_ASCII_SPACE); 										/* remove SPACES at the beginn and at the End */
		cmd_interface_table[ui_static_cmd_counter].cmd=cmd_string;				/* add command-string to table */
		cmd_interface_table[ui_static_cmd_counter].cmd_func_ptr=cmd_func_ptr;	/* add function-pointer to table */
		ui_static_cmd_counter++; /* inkrement command-counter */
		ret_val =0; /* clear preset errror */
	}

	return (ret_val);
}/* cmd_interface_add_cmd */

/**
 	@brief		routine called to execute command interface
 	@param		*string - to attach at the actual command_line buffer
	@return		The content of the return-value have been split into two parts
	@return		The three lowest bits are used as counter for executed commands
	@return		The remaining "upper" bits represent bit-coded errors,
	@return		due to the possibilitiy of multiple errors per call.
 	@return		  <0 		- error occured
 	@return		#1000 1XXX	- #84	- operands have to many characters
 	@return		#1001 0XXX	- #90	- too many operands in the command-line
 	@return		#1010 0XXX	- #A0	- no valid command found
 	@return		#1100 0XXX	- #B0	- too many signs in comand-line
 	@return		  0			- no end of command-line detects
 	@return		 0> 		- number of commands which have been executed
 	@details	This routine has to be called to execute the command-interface
 				It attaches the forwarded string to the internal command-line buffer
 				and its sub-functions interprets the line before it executes the corresponding
 				command-function given by the user
 */

int8_t execute_cmd_interface(char *string){
	/* local variables */
	register uint8_t	ui_cnt		=0;
	register char		act_char	=1;
	register int8_t		cmd_ret_val	=0;
	register uint8_t	cmd_count	=0;
	int8_t 				ret_val		=0;
	/* start reading string */
	act_char=string[ui_cnt++];	/* Read next sign */

	/* adding string to cmd_line while looking for an EOL */
	while	( 		(cmd_line_act_length<_CMD_INTERFACE_MAX_LENGTH)
				&&	(act_char!=0)									/* EOS ? */
	){
		if(_CMD_INTERFACE_EOL==act_char){		/* End of commandline ? */
			cmd_line[cmd_line_act_length]=0; 	/* Set EOS in string of commandline */
			cmd_ret_val=execute_command();
			switch (cmd_ret_val) {
				case  0:						/* set error: no valid command */
						ret_val |= _CMD_INTERFACE_ERR_NO_CMD;
						break;
				case -1:						/* set error: operands have to many characters */
						ret_val |= _CMD_INTERFACE_ERR_OP_CHAR;
						break;
				case -2:						/* set error: too many operands in the command-line  */
						ret_val |= _CMD_INTERFACE_ERR_OP_2_MANY;
						break;
				default:						/* count up counter for executed commands */
						if (0<cmd_ret_val)	cmd_count++;
						break;
			} /* switch (cmd_ret_val)*/
		} else if (_ASCII_BS==act_char){
			cmd_line_act_length--;
		} else {
			cmd_line[cmd_line_act_length++]=act_char;
		}
		act_char=string[ui_cnt++];		/* Read next sign */
	} /* while */

	/* to many signs in commandline ? */
	if (cmd_line_act_length>=_CMD_INTERFACE_MAX_LENGTH){
		cmd_line_act_length=0;
		/* set error: to many signs in commandline */
		ret_val |= _CMD_INTERFACE_ERR_CMD_LINE;
	}

	/* Set EOS in string of commandline */
	cmd_line[cmd_line_act_length]=0;

	/* combine ret_val-information */
	ret_val |= cmd_count& ~_CMD_INTERFACE_ERR_MASK;

	return(ret_val);
} /*execute_cmd_interface()*/


/**
 	@brief		internal function which executes the commands
 	@param		none
 	@return		<0 	- error
	@return		-1	- operands have to many characters	(these are the ret_vals of split_command_line)
 	@return		-2	- too many operands in the command-line (these are the ret_vals of split_command_line)
	@return		 0 	- no valid command
	@return		 1	- a command hast been successfully executed.
 	@details    this functions checks whether the given command exists in the command-table
 				if it exists it executes the function referenced in the function-pointer
 				which belongs the specific command.
 */
int8_t execute_command(){
	register uint8_t ui_cnt=0;
	register int8_t  ret_val;
	/* Analyze command_line and transfer segmented data into data-array */
		ret_val=split_command_line();
		if (0>ret_val) goto func_end;

	/* check whether the command is valid and execute the corresponding function */
		ret_val=0;
		for(ui_cnt=0; _CMD_INTERFACE_NO_OF_COMMANDS>ui_cnt; ui_cnt++){
			if ( cmd_interface_table[ui_cnt].cmd_func_ptr != NULL )  {
				if ( strcmp( cmd_interface_table[ui_cnt].cmd,&cmd_line_split_parts[0][0] )==0 ) {
					cmd_interface_table[ui_cnt].cmd_func_ptr(&cmd_line_split_parts[1][0], &cmd_line_split_parts[2][0],&cmd_line_split_parts[3][0],&cmd_line_split_parts[4][0]);
					ui_cnt = _CMD_INTERFACE_NO_OF_COMMANDS; /* skip rest - only one command is possible per line */
					ret_val = 1; /* one cmd executed */
				}
			} /* if ( cmd_interface_table[ui_cnt].cmd_func_ptr != NULL ) */
		}

func_end:
	/* clean command line */
		cmd_line_act_length			=0;

	return(ret_val);
}/*inline int8_t execute_command()*/



/**
 	@brief		internal function which splits the act. command line into separated components, which are
 				separated by split signs.
 	@param		none
 	@return		<0 	- 	error
 	@return		-1	-	an operand has to many characters
 	@return		-2	-	too many operands in the command-line
 	@return		>0 	- 	number of strings/operands which have been found, means separated by split-signs.
 */
int8_t split_command_line(){
	register int8_t	sign_cnt	=0;	/* Continuous sign counter */
	register int8_t	op_cnt		=0;	/* operand counter which shall not exceed cmd_interface_max_no_operands*/
	register int8_t split_pos	=0;	/* memory for act / last position of a split sign */

	/* whipe previous results */
	for (op_cnt=0; op_cnt < _CMD_INTERFACE_MAX_NO_OPERANDS; op_cnt++) cmd_line_split_parts[op_cnt][0]=0; /* Set EOS */

	/* check for command */
	op_cnt=0;
	trim(&cmd_line[0],_ASCII_SPACE);
	split_pos = split_string(&cmd_line[0],&cmd_line_split_parts[op_cnt][0],_CMD_INTERFACE_COMMAND_SPLIT_SIGN,_CMD_INTERFACE_MAX_OPERAND_LENGTH);
	sign_cnt=split_pos; 										/* set pos for next search */
	trim(&cmd_line_split_parts[op_cnt][0],_ASCII_SPACE);

	/* check for operands */
	while( 		(sign_cnt <  cmd_line_act_length)				/* do not exceed act. length of command line */
			&&  (split_pos > 0					)				/* still something to find?  */
	){
		/* increment counter for operands */
		if (_CMD_INTERFACE_MAX_NO_OPERANDS>(++op_cnt)) { 		/* if there are still operands possible.. then go on */
			split_pos = split_string(&cmd_line[sign_cnt],&cmd_line_split_parts[op_cnt][0],_CMD_INTERFACE_OPERAND_SPLIT_SIGN,_CMD_INTERFACE_MAX_OPERAND_LENGTH);
			if (split_pos>0){
				trim(&cmd_line_split_parts[op_cnt][0],_ASCII_SPACE);		/* remove leading and tailing SPACES */
				sign_cnt+=split_pos; 							/* set pos for next search */
			}
		} else { 												/* no more operands are allowed */
			split_pos = 0;
		}
	} /* while */

	/* copy possible remaining parts if operands are still allowed to add */
	if (_CMD_INTERFACE_MAX_NO_OPERANDS>op_cnt) {
		if (split_pos==0) {
			if (0<=string_copy(&cmd_line[sign_cnt], &cmd_line_split_parts[op_cnt][0], _CMD_INTERFACE_MAX_OPERAND_LENGTH)) {
				trim(&cmd_line_split_parts[op_cnt][0],_ASCII_SPACE);		/* remove leading and tailing SPACES */
			} else {
				split_pos = -1; /* too many characters per operand */
			}
		}
	} else op_cnt = -2; /* error - too many operands */

	if (-1 == split_pos ) op_cnt=-1; /* error - too many char for one operand */

return(op_cnt);
}/*inline int8_t execute_command()*/




#undef  CMD_INTERFACE_C


/** STORAGE-PLACE
		// simulate command 1
		sprintf(str_rcv_usart,"cmd  test 1 , test2,test3,test4     %c",toascii(13));
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"cmd  was 1 , param2, p a r a m 3 , p4, p5 %c",toascii(13));
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"cmd  op1 , op2,op3 %c",toascii(13));
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"a op1,op2,op3 %c b test, was, das, wird%c",toascii(13),toascii(13));
 		iCnt=execute_cmd_interface(str_rcv_usart);

		sprintf(str_rcv_usart,"c");
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"m");
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"d");
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart," ");
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"p1");
		iCnt=execute_cmd_interface(str_rcv_usart);
		sprintf(str_rcv_usart,"%c",toascii(13));
		iCnt=execute_cmd_interface(str_rcv_usart);


	// simulate wrong commands and errors
	sprintf(str_rcv_usart,"bloedsin so , doof%c  a 1,2,3,4,5%c a 1,2,3,4%c",toascii(13),toascii(13),toascii(13));
	iCnt=execute_cmd_interface(str_rcv_usart);

	sprintf(str_rcv_usart,"bloedsin so , doof%c",toascii(13));
	iCnt=execute_cmd_interface(str_rcv_usart);
	sprintf(str_rcv_usart,"so ein,scheiss,was,soll, das%c",toascii(13));
	iCnt=execute_cmd_interface(str_rcv_usart);

	sprintf(str_rcv_usart,"das ist ein kleiner Test %cab was dies wohl bewirken %cwird%c",toascii(13),toascii(13),toascii(13));
	iCnt=execute_cmd_interface(str_rcv_usart);



*/
